﻿using System;
using System.Data;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.ComponentModel;
using System.Threading.Tasks;
using System.Collections.Generic;
using System.Collections.Specialized;

namespace Local_Backup_Utility
{
    public partial class FormDirectories : Form
    {
        public FormDirectories()
        {
            InitializeComponent();
        }

        private void FormDirectories_Load(object sender, EventArgs e)
        {
            if (Properties.Settings.Default.theme == "dark")
            {
                this.BackColor = Color.FromArgb(26, 26, 27);
                Themes.changeTheme("dark", this.Controls);
            }

            for (int i = 0; i < FormMain.sources.Count(); i++)
            {
                ListViewItem listItem = new ListViewItem(FormMain.sources[i]);
                listItem.SubItems.Add(FormMain.destinations[i]);
                listDirectories.Items.Add(listItem);
            }
        }

        private void buttonSource_Click(object sender, EventArgs e)
        {
            if (folderBrowserDialog.ShowDialog() == DialogResult.OK) textSource.Text = folderBrowserDialog.SelectedPath;
        }

        private void buttonDestination_Click(object sender, EventArgs e)
        {
            if (folderBrowserDialog.ShowDialog() == DialogResult.OK) textDestination.Text = folderBrowserDialog.SelectedPath;
        }

        private void buttonAdd_Click(object sender, EventArgs e)
        {
            if (textSource.Text == string.Empty || textDestination.Text == string.Empty)
            {
                MessageBox.Show("Directories not selected");
                return;
            }

            string[] newList = new string[] { "(null)" };

            FormMain.sources.Add(textSource.Text);
            FormMain.destinations.Add(textDestination.Text);
            FormMain.exceptions = Dataset.Add(FormMain.exceptions, newList.Cast<string>().ToList());

            Dataset.SaveAll();
            updateDirectories();

            textSource.Text = string.Empty;
            textDestination.Text = string.Empty;
        }

        private void buttonRemove_Click(object sender, EventArgs e)
        {
            if(listDirectories.SelectedItems.Count != 1)
            {
                MessageBox.Show("Item to remove not selected");
                return;
            }

            int index = listDirectories.Items.IndexOf(listDirectories.SelectedItems[0]);

            FormMain.sources.RemoveAt(index);
            FormMain.destinations.RemoveAt(index);
            FormMain.exceptions = Dataset.Remove(FormMain.exceptions, index);

            Dataset.SaveAll();
            updateDirectories();
        }

        private void buttonExceptions_Click(object sender, EventArgs e)
        {
            if (listDirectories.SelectedItems.Count != 1)
            {
                MessageBox.Show("Item to manage not selected");
                return;
            }

            Form form = new FormExceptions(listDirectories.Items.IndexOf(listDirectories.SelectedItems[0]));
            form.ShowDialog();
        }

        void updateDirectories()
        {
            listDirectories.Items.Clear();

            for (int i = 0; i < FormMain.sources.Count(); i++)
            {
                ListViewItem listItem = new ListViewItem(FormMain.sources[i]);
                listItem.SubItems.Add(FormMain.destinations[i]);
                listDirectories.Items.Add(listItem);
            }
        }

        private void listDirectories_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Delete) buttonRemove.PerformClick();
            e.Handled = true;
        }

        private void textSource_Click(object sender, EventArgs e)
        {
            buttonSource.PerformClick();
        }

        private void textDestination_Click(object sender, EventArgs e)
        {
            buttonDestination.PerformClick();
        }
        private void listDirectories_DoubleClick(object sender, EventArgs e)
        {
            buttonExceptions.PerformClick();
        }
    }
}
