﻿using System;
using System.IO;
using System.Net;
using System.Data;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Threading;
using System.Management;
using System.Diagnostics;
using System.Net.Sockets;
using System.Windows.Forms;
using System.ComponentModel;
using System.Threading.Tasks;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Collections.Specialized;

namespace Local_Backup_Utility
{
    public partial class FormMain : Form
    {
        bool debugMode = false;
        public static float versionLocalBackupUtility = 1.1f;

        public static List<string> sources = new List<string>();
        public static List<string> destinations = new List<string>();
        public static List<string>[] exceptions;

        List<string> sourceDirectories = new List<string>();
        List<string> destinationDirectories = new List<string>();
        string _removed;

        FormWarnings formWarnings;
        public static List<string>[] warnings = new List<string>[2];
        public static List<string>[] history = new List<string>[2];

        List<string> skip = new List<string>();

        int stage;
        string masterDirectory;
        List<string> directoriesRemoved = new List<string>();
        int totalFilesProcessed;
        int totalFilesFound;
        int reverseTotalFilesProcessed;
        int reverseTotalFilesFound;
        int reverseTotalDirectoriesProcessed;
        int copiedFiles;
        int removedFiles;
        int renamedFiles;
        int reverseRenamedFiles;        

        DateTime nextUpdateDate = new DateTime();
        string logFilePath = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments) + "\\" + "Local Backup Utility.txt";
        string[] daysOfWeek = new string[] { "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" };
        public static bool closeForm = false;

        System.Threading.Timer _updateTimer;
        System.Threading.Timer _threeDotTimer;
        Thread _showLoop;
        Thread _informProgress;
        bool _updateTimerRunning = false;

        Color colorRed = Color.DarkRed;
        Color colorGreen = Color.DarkGreen;
        Color colorBlack = SystemColors.ControlText;

        string[] excludedDirectories = new string[] {
            "$RECYCLE.BIN",
            "System Volume Information"
        };

        string[] excludedFiles = new string[] {
            "Thumbs.db"
        };

        public FormMain()
        {
            InitializeComponent();

            warnings[0] = new List<string>();
            warnings[1] = new List<string>();

            history[0] = new List<string>();
            history[1] = new List<string>();
        }

        private void FormMain_Load(object sender, EventArgs e)
        {
            Process[] processes = Process.GetProcessesByName(Process.GetCurrentProcess().ProcessName);

            if (processes.Length > 1)
            {
                string[] args = Environment.GetCommandLineArgs();
                if(args.Contains("restart") == false)
                {
                    closeForm = true;
                    Thread.Sleep(1000);
                    this.Close();
                    return;
                }
            }

            if (Properties.Settings.Default.theme == "dark")
            {
                this.BackColor = Color.FromArgb(26, 26, 27);
                Themes.changeTheme("dark", this.Controls);

                colorRed = Color.Red;
                colorGreen = Color.LightGreen;
                colorBlack = Color.White;
            }

            if (Properties.Settings.Default.sources != null)
            {
                sources = Properties.Settings.Default.sources.Cast<string>().ToList();
                destinations = Properties.Settings.Default.destinations.Cast<string>().ToList();
                exceptions = Dataset.Unpack(Properties.Settings.Default.exceptions.Cast<string>().ToArray());
            }

            if (Properties.Settings.Default.historyTime != null)
            {
                history[0] = Properties.Settings.Default.historyTime.Cast<string>().ToList();
                history[1] = Properties.Settings.Default.historyMessage.Cast<string>().ToList();

                string[] time = FormMain.history[0].ToArray();
                string[] message = FormMain.history[1].ToArray();

                for (int i = time.Length - 1; i >= 0; i--)
                {
                    ListViewItem listItem = new ListViewItem(time[i]);
                    listItem.SubItems.Add(message[i]);
                    listActions.Items.Add(listItem);
                }
            }

            updateDirectories();
            updateGlobalSettings();

            if (Properties.Settings.Default.schedule[0] == "enabled" && sources.Count != 0)
            {
                _updateTimer = new System.Threading.Timer(updateTimer, null, 0, 1000);
                _updateTimerRunning = true;
            }

            notifyIcon.ContextMenuStrip = new System.Windows.Forms.ContextMenuStrip();
            notifyIcon.ContextMenuStrip.Items.Add("Local Backup Utility", null, this.notifyIcon_MouseDoubleClick);
            notifyIcon.ContextMenuStrip.Items.Add("-");
            notifyIcon.ContextMenuStrip.Items.Add("Exit Local Backup Utility", null, exitToolStripMenuItem_Click);
            notifyIcon.Visible = true;

            _showLoop = new Thread(showLoop);
            _showLoop.Start();
        }

        private void FormMain_Shown(object sender, EventArgs e)
        {
            if (Properties.Settings.Default.startMinimised == true) Hide();

            this.Opacity = 100;
        }

        private void notifyIcon_MouseDoubleClick(object sender, EventArgs e)
        {
            Show();
            this.WindowState = FormWindowState.Normal;
        }

        void showLoop()
        {
            while(true)
            { 
                Process[] processes = Process.GetProcessesByName(Process.GetCurrentProcess().ProcessName);

                if (processes.Length > 1)
                {
                    for(int i = 0; i < processes.Length; i++)
                    {
                        if (processes[i].Id == Process.GetCurrentProcess().Id) continue;

                        processes[i].Kill();

                        this.Invoke((MethodInvoker)delegate
                        {
                            Show();
                            this.WindowState = FormWindowState.Normal;
                        });
                    }
                }

                Thread.Sleep(500);
            }
        }

        private void FormMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (closeForm == false)
            {
                e.Cancel = true;

                this.WindowState = FormWindowState.Minimized;
                notifyIcon.Visible = true;

                Hide();

                notifyIcon.ShowBalloonTip(5000, "Minimised to system tray", "Local Backup Utility has been minimised to the system tray", ToolTipIcon.None);
            }
            else if(closeForm == true)
            {
                _showLoop.Abort();
            }
        }

        private void defaultSettingsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if(buttonBackupNow.Enabled == false)
            {
                MessageBox.Show("Backup in progress" + Environment.NewLine + "Please abort backup before resetting");
                return; 
            }

            DialogResult checkReset = MessageBox.Show("All settings will be reset to default and Local Backup Utility will restart" + Environment.NewLine + "Do you wish to continue?", "Reset settings and restart", MessageBoxButtons.YesNo);

            if (checkReset == DialogResult.Yes)
            {
                Properties.Settings.Default.Reset();

                Process.Start(Application.ExecutablePath, "restart");

                closeForm = true;
                this.Close();
            }
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (buttonBackupNow.Enabled == false)
            {
                MessageBox.Show("Backup in progress" + Environment.NewLine + "Please abort backup before exiting");
                return;
            }

            closeForm = true;
            this.Close();
        }

        private void buttonBackupNow_Click(object sender, EventArgs e)
        {
            if(sources.Count == 0)
            {
                MessageBox.Show("Directories have not been established" + Environment.NewLine + "Click the Manage Directories button");
                return;
            }

            Thread thread = new Thread(() => backup(true));
            thread.Start();
        }

        private void buttonAbort_Click(object sender, EventArgs e)
        {
            buttonAbort.Enabled = false;
            if (Properties.Settings.Default.theme == "dark") Themes.changeTheme("dark", this.Controls);

            labelStatus.ForeColor = colorRed;
            labelThreeDot.ForeColor = colorRed;

            labelStatus.Text = "Backup will abort after current file has finished copying";
            labelStatus.Update();
        }

        void backup(bool fullBackup)
        {
            /* ----- CREATE CURRENT COPIES OF VARIABLES ----- */

            string _updated = Properties.Settings.Default.updated;
            _removed = Properties.Settings.Default.removed;

            List<string> _sources = sources;
            List<string> _destinations = destinations;
            List<string>[] _exceptions = exceptions;

            /* ----- CREATE CURRENT COPIES OF VARIABLES COMPLETE ----- */

            buttonBackupNow.Invoke((MethodInvoker)delegate { buttonBackupNow.Enabled = false; });
            buttonAbort.Invoke((MethodInvoker)delegate { buttonAbort.Enabled = true; });
            if (Properties.Settings.Default.theme == "dark") this.Invoke((MethodInvoker)delegate { Themes.changeTheme("dark", this.Controls); });
            skip.Clear();

            _threeDotTimer = new System.Threading.Timer(threeDotTimer, null, 0, 333);
            if (_updateTimerRunning == true)
            {
                _updateTimer.Dispose();
                _updateTimerRunning = false;
            }
            labelStatus.Invoke((MethodInvoker)delegate{ labelStatus.ForeColor = colorGreen; });
            labelThreeDot.Invoke((MethodInvoker)delegate{ labelThreeDot.ForeColor = colorGreen; });

            labelNextBackup.Invoke((MethodInvoker)delegate { labelNextBackup.Text = "Next backup: In progress..."; });
            labelNextBackup.Invoke((MethodInvoker)delegate { labelNextBackup.Update(); });

            int iStart = 0, iEnd = _sources.Count();
            if(fullBackup == false)
            {
                iStart = listDirectories.FocusedItem.Index;
                iEnd = iStart + 1;
            }

            for (int i = iStart; i < iEnd; i++)
            {
                /* ----- INFORM ACTION ----- */

                DateTime currentDateTime = DateTime.Now;
                string formattedDateTime = currentDateTime.ToString("dd/MM/yyyy - HH:mm:ss");

                ListViewItem listItem = new ListViewItem(formattedDateTime);
                listItem.SubItems.Add("Task " + (i + 1) + " Working... " + _sources[i] + " to " + _destinations[i]);
                listActions.Invoke((MethodInvoker)delegate { listActions.Items.Insert(0, listItem); });

                listActions.Invoke((MethodInvoker)delegate { listActions.Update(); });

                /* ----- INFORM ACTION COMPLETE ----- */



                /* ----- BACKUP ----- */

                Stopwatch timer = new Stopwatch();
                timer.Start();

                _informProgress = new Thread(informProgress);
                _informProgress.Start();

                masterDirectory = _sources[i];
                directoriesRemoved.Clear();
                totalFilesFound = 0;
                totalFilesProcessed = 0;
                reverseTotalFilesFound = 0;
                reverseTotalFilesProcessed = 0;
                reverseTotalDirectoriesProcessed = 0;
                copiedFiles = 0;
                removedFiles = 0;
                renamedFiles = 0;
                reverseRenamedFiles = 0;

                stage = 0;
                sourceDirectories.Clear();
                destinationDirectories.Clear();
                traverseDirectories(_sources[i], _destinations[i], _exceptions[i]);

                stage = 1;
                processFilesInDirectory(_updated, _exceptions[i]);

                stage = 2;
                sourceDirectories.Clear();
                destinationDirectories.Clear();
                reverseTraverseDirectories(_sources[i], _destinations[i]);

                stage = 3;
                reverseProcessFilesInDirectory(_removed, _exceptions[i]);

                stage = 4;
                removeDirectories(_removed);

                _informProgress.Abort();
                timer.Stop();

                /* ----- BACKUP COMPLETE ----- */



                /* ----- LOG ACTION ----- */

                string time;
                if (timer.Elapsed.Minutes == 0 && timer.Elapsed.Seconds == 0) time = "(" + timer.Elapsed.Milliseconds + "ms)";
                else if (timer.Elapsed.Minutes == 0) time = "(" + timer.Elapsed.Seconds + "s " + timer.Elapsed.Milliseconds + "ms)";
                else time = "(" + timer.Elapsed.Minutes + "m " + timer.Elapsed.Seconds + "s " + timer.Elapsed.Milliseconds + "ms)";

                currentDateTime = DateTime.Now;
                formattedDateTime = currentDateTime.ToString("dd/MM/yyyy - HH:mm:ss");
                
                listActions.Invoke((MethodInvoker)delegate { listActions.Items.RemoveAt(0); });
                listItem = new ListViewItem(formattedDateTime);

                /* > FORM LOG MESSAGE */

                string logMessage = "Task " + (i + 1);
                logMessage += (buttonAbort.Enabled == true ? " Complete. " : " Aborted. ");
                logMessage += totalFilesFound + " " + (totalFilesFound == 1 ? "file" : "files") + " found. " + copiedFiles + " " + (copiedFiles == 1 ? "file" : "files") + " copied. ";
                if (removedFiles > 0) logMessage += removedFiles + " " + (removedFiles == 1 ? "file/folder" : "files/folders") + " removed. ";
                if (renamedFiles > 0) logMessage += renamedFiles + " " + (renamedFiles == 1 ? "file/folder" : "files/folders") + " renamed. ";
                logMessage += time;

                /* > FORM LOG MESSAGE COMPLETE */

                listItem.SubItems.Add(logMessage);
                listActions.Invoke((MethodInvoker)delegate { listActions.Items.Insert(0, listItem); });
                
                history[0].Add(formattedDateTime);
                history[1].Add(logMessage);

                logLog(logMessage);

                StringCollection timeCollection = new StringCollection();
                timeCollection.AddRange(history[0].ToArray());
                Properties.Settings.Default.historyTime = timeCollection;

                StringCollection messageCollection = new StringCollection();
                messageCollection.AddRange(history[1].ToArray());
                Properties.Settings.Default.historyMessage = messageCollection;

                Properties.Settings.Default.Save();
                
                listActions.Invoke((MethodInvoker)delegate { listActions.Update(); });

                /* ----- LOG ACTION COMPLETE ----- */
            }
            
            labelStatus.Invoke((MethodInvoker)delegate { labelStatus.Text = "Not started"; });

            _threeDotTimer.Dispose();
            labelThreeDot.Invoke((MethodInvoker)delegate { labelThreeDot.Text = "Status:"; });
            labelStatus.Invoke((MethodInvoker)delegate { labelStatus.ForeColor = colorBlack; });
            labelThreeDot.Invoke((MethodInvoker)delegate { labelThreeDot.ForeColor = colorBlack; });

            if (Properties.Settings.Default.schedule[0] == "enabled" && _sources.Count != 0)
            {
                _updateTimer = new System.Threading.Timer(updateTimer, null, 1000, 1000);
                _updateTimerRunning = true;
            }
            else labelNextBackup.Invoke((MethodInvoker)delegate { labelNextBackup.Text = "Next backup: Not scheduled"; });

            buttonBackupNow.Invoke((MethodInvoker)delegate { buttonBackupNow.Enabled = true; });
            buttonAbort.Invoke((MethodInvoker)delegate { buttonAbort.Enabled = false; });
            if (Properties.Settings.Default.theme == "dark") this.Invoke((MethodInvoker)delegate { Themes.changeTheme("dark", this.Controls); });
        }

        void traverseDirectories(string source, string destination, List<string> exceptionsList)
        {
            if (buttonAbort.Enabled == false) return;

            string sourceDirectoryName = new DirectoryInfo(source).Name;
            for (int i = 0; i < excludedDirectories.Length; i++) if (sourceDirectoryName == excludedDirectories[i]) return;

            try
            {
                string[] sourceFiles = Directory.GetFiles(source);
                totalFilesFound += sourceFiles.Length;
                for (int i = 0; i < excludedFiles.Length; i++) if (sourceFiles.Contains(source + "\\" + excludedFiles[i]) == true) totalFilesFound--;

                sourceDirectories.Add(source);
                destinationDirectories.Add(destination);

                string[] sourceSubDirectories = Directory.GetDirectories(source);
                foreach (string sourceSubDirectory in sourceSubDirectories)
                {
                    if (exceptionsList.Contains(sourceSubDirectory)) continue;

                    string subdirName = new DirectoryInfo(sourceSubDirectory).Name;
                    string destinationSubDirectory = destination + "\\" + subdirName;

                    traverseDirectories(sourceSubDirectory, destinationSubDirectory, exceptionsList);

                    if (debugMode == true) Thread.Sleep(500);
                }
            }
            catch (Exception ex)
            {
                processException(source, destination, ex);
            }
        }

        void reverseTraverseDirectories(string source, string destination)
        {
            if (buttonAbort.Enabled == false) return;
            if (skip.Contains(destination)) return;

            string destinationDirectoryName = new DirectoryInfo(destination).Name;
            for (int i = 0; i < excludedDirectories.Length; i++) if (destinationDirectoryName == excludedDirectories[i]) return;

            try
            {
                if (destination.Contains("{old LBU}") == true) return;
                if (destination.Contains("{removed LBU}") == true) return;

                if (Directory.Exists(destination) == true && Directory.Exists(source) == false)
                {
                    if (Directory.Exists(masterDirectory) == true)
                    {
                        directoriesRemoved.Add(destination);
                        return;
                    }
                }

                string[] destinationFiles = Directory.GetFiles(destination);
                reverseTotalFilesFound += destinationFiles.Length;
                for (int i = 0; i < excludedFiles.Length; i++) if (destinationFiles.Contains(destination + "\\" + excludedFiles[i]) == true) reverseTotalFilesFound--;

                sourceDirectories.Add(source);
                destinationDirectories.Add(destination);

                string[] destinationSubDirectories = Directory.GetDirectories(destination);
                foreach (string destinationSubDirectory in destinationSubDirectories)
                {
                    string subdirName = new DirectoryInfo(destinationSubDirectory).Name;
                    string sourceSubDirectory = source + "\\" + subdirName;

                    reverseTraverseDirectories(sourceSubDirectory, destinationSubDirectory);

                    if (debugMode == true) Thread.Sleep(500);
                }
            }
            catch (Exception ex)
            {
                processException(source, destination, ex);
            }
        }

        void processFilesInDirectory(string updated, List<string> exceptionsList)
        {
            for (int i = 0; i < sourceDirectories.Count; i++)
            {
                if (exceptionsList.Contains(sourceDirectories[i])) continue;

                string[] sourceFiles;
                try
                {
                    sourceFiles = Directory.GetFiles(sourceDirectories[i]);

                    if (Directory.Exists(destinationDirectories[i]) == false) Directory.CreateDirectory(destinationDirectories[i]);
                }
                catch (Exception ex)
                {
                    processException(sourceDirectories[i], destinationDirectories[i], ex);
                    continue;
                }

                foreach(string sourceFile in sourceFiles)
                {
                    try
                    {
                        if (Directory.Exists(sourceDirectories[i]) == false) throw new FileNotFoundException("Could not find a part of the path '" + sourceDirectories[i] + "'.", sourceDirectories[i]);

                        FileInfo sourceFileInfo = new FileInfo(sourceFile);
                        string fileName = sourceFileInfo.Name;
                        string destinationFile = destinationDirectories[i] + "\\" + fileName;

                        bool cont = false;
                        for (int a = 0; a < excludedFiles.Length; a++) if (fileName == excludedFiles[a]) cont = true;
                        if (cont == true) continue;

                        totalFilesProcessed++;

                        if (File.Exists(destinationFile) == true)
                        {
                            FileInfo destinationFileInfo = new FileInfo(destinationFile);

                            if (sourceFileInfo.LastWriteTime > destinationFileInfo.LastWriteTime)
                            {
                                if (updated == "overwrite")
                                {
                                    File.Copy(sourceFile, destinationFile, true);

                                    logLog("Newer file exists. File overwritten: " + destinationFile);
                                }
                                else if (updated == "rename")
                                {
                                    int a = 0;

                                    string renameName = Path.GetFileNameWithoutExtension(destinationFile);
                                    string renameExtension = Path.GetExtension(destinationFile);
                                    string destinationFileRenamed = destinationDirectories[i] + @"\" + renameName + " {old LBU} {" + a + "}" + renameExtension;

                                    while (File.Exists(destinationFileRenamed) == true)
                                    {
                                        a++;
                                        destinationFileRenamed = destinationDirectories[i] + @"\" + renameName + " {old LBU} {" + a + "}" + renameExtension;
                                    }

                                    File.Move(destinationFile, destinationFileRenamed);
                                    File.Copy(sourceFile, destinationFile);

                                    renamedFiles++;

                                    logLog("Newer file exists. File renamed: " + destinationFileRenamed);
                                }

                                copiedFiles++;
                            }
                        }
                        else if (File.Exists(destinationFile) == false)
                        {
                            File.Copy(sourceFile, destinationFile);
                            copiedFiles++;
                        }

                        if (debugMode == true) Thread.Sleep(500);
                    }
                    catch(Exception ex)
                    {
                        processException(sourceDirectories[i], destinationDirectories[i], ex);
                        break;
                    }

                    if (buttonAbort.Enabled == false) return;
                }
            }
        }

        void reverseProcessFilesInDirectory(string removed, List<string> exceptionsList)
        {
            for (int i = 0; i < destinationDirectories.Count; i++)
            {
                if (skip.Contains(destinationDirectories[i])) continue;

                string[] destinationFiles;
                try
                {
                    destinationFiles = Directory.GetFiles(destinationDirectories[i]);
                }
                catch (Exception ex)
                {
                    processException(sourceDirectories[i], destinationDirectories[i], ex);
                    continue;
                }

                foreach (string destinationFile in destinationFiles)
                {
                    try
                    {
                        if (Directory.Exists(sourceDirectories[i]) == false) throw new FileNotFoundException("Could not find a part of the path '" + sourceDirectories[i] + "'.", sourceDirectories[i]);

                        if (destinationFile.Contains("{old LBU}") == true) continue;
                        if (destinationFile.Contains("{removed LBU}") == true) continue;

                        string fileName = new FileInfo(destinationFile).Name;
                        string sourceFile = sourceDirectories[i] + "\\" + fileName;

                        bool cont = false;
                        for (int a = 0; a < excludedFiles.Length; a++) if (fileName == excludedFiles[a]) cont = true;
                        if (cont == true) continue;

                        reverseTotalFilesProcessed++;

                        if (File.Exists(sourceFile) == false)
                        {
                            if (removed == "delete")
                            {
                                File.Delete(destinationFile);
                                removedFiles++;

                                logLog("File no longer exists. File removed: " + destinationFile);
                            }
                            else if (removed == "rename")
                            {
                                int a = 0;

                                string renameName = Path.GetFileNameWithoutExtension(destinationFile);
                                string renameExtension = Path.GetExtension(destinationFile);
                                string destinationFileRenamed = destinationDirectories[i] + "\\" + renameName + " {removed LBU} {" + a + "}" + renameExtension;

                                while (File.Exists(destinationFileRenamed) == true)
                                {
                                    a++;
                                    destinationFileRenamed = destinationDirectories[i] + "\\" + renameName + " {removed LBU} {" + a + "}" + renameExtension;
                                }

                                File.Move(destinationFile, destinationFileRenamed);
                                renamedFiles++;
                                reverseRenamedFiles++;

                                logLog("File no longer exists. File renamed: " + destinationFileRenamed);
                            }
                        }

                        if (debugMode == true) Thread.Sleep(500);
                    }
                    catch (Exception ex)
                    {
                        processException(sourceDirectories[i], destinationDirectories[i], ex);
                        break;
                    }

                    if (buttonAbort.Enabled == false) return;
                }
            }
        }

        void removeDirectories(string removed)
        {
            for (int i = 0; i < directoriesRemoved.Count; i++)
            {
                try
                {
                    if(removed == "delete")
                    {
                        Directory.Delete(directoriesRemoved[i], true);
                        removedFiles++;

                        logLog("Folder no longer exists. Folder removed: " + directoriesRemoved[i]);
                    }
                    else if (removed == "rename")
                    {
                        int a = 0;
                        string destinationDirectoryRenamed = directoriesRemoved[i] + " {removed LBU} {" + a + "}";
                        destinationDirectoryRenamed = directoriesRemoved[i] + " {removed LBU} {" + a + "}";

                        while (Directory.Exists(destinationDirectoryRenamed) == true)
                        {
                            a++;
                            destinationDirectoryRenamed = directoriesRemoved[i] + " {removed LBU} {" + a + "}";
                        }

                        Directory.Move(directoriesRemoved[i], destinationDirectoryRenamed);
                        renamedFiles++;

                        logLog("Folder no longer exists. Folder renamed: " + destinationDirectoryRenamed);
                    }                    
                }
                catch (Exception ex)
                {
                    processException(sourceDirectories[i], destinationDirectories[i], ex);
                    break;
                }

                reverseTotalDirectoriesProcessed++;

                if(debugMode == true) Thread.Sleep(500);
            }
        }

        void processException(string source, string destination, Exception ex)
        {
            if (Directory.Exists(source) == false || Directory.Exists(destination) == false)
            {
                string warningMessage = String.Empty;

                if (Directory.Exists(source) == false) warningMessage = "Access to path '" + source + "' lost.  Skipping.";
                else if (Directory.Exists(destination) == false) warningMessage = "Access to path '" + destination + "' lost.  Skipping.";

                logLog(warningMessage);
                logWarning(warningMessage);
            }
            else
            {
                logLog(ex.Message);
                logWarning(ex.Message);
            }

            skip.Add(destination);
        }

        void informProgress()
        {
            while (true)
            {
                string message = string.Empty;

                if (stage == 0)
                {
                    message = "Finding directories... " + sourceDirectories.Count + " " + (sourceDirectories.Count == 1 ? "directory" : "directories") + " found";
                }
                else if (stage == 1)
                {
                    message = "Copying files... " + copiedFiles + " " + (copiedFiles == 1 ? "file" : "files") + " copied";
                    if (renamedFiles > 0) message += ", " + renamedFiles + " " + (renamedFiles == 1 ? "file" : "files") + " renamed";
                    message += " (" + totalFilesProcessed + "/" + totalFilesFound + ")";
                }
                else if (stage == 2)
                {
                    message = "Cleaning up... " + destinationDirectories.Count + " " + (destinationDirectories.Count == 1 ? "directory" : "directories") + " found";
                }
                else if (stage == 3)
                {
                    message = "Cleaning up... ";

                    if (_removed == "delete") message += removedFiles + " " + (removedFiles == 1 ? "file" : "files") + " removed.";
                    else if (_removed == "rename") message += reverseRenamedFiles + " " + (reverseRenamedFiles == 1 ? "file" : "files") + " renamed.";

                    message += " (" + reverseTotalFilesProcessed + "/" + reverseTotalFilesFound + ")";
                }
                else if (stage == 4)
                {
                    message = "Cleaning up... ";

                    if (_removed == "delete") message += reverseTotalDirectoriesProcessed + " " + (reverseTotalDirectoriesProcessed == 1 ? "directory" : "directories") + " removed.";
                    else if (_removed == "rename") message += reverseTotalDirectoriesProcessed + " " + (reverseTotalDirectoriesProcessed == 1 ? "directory" : "directories") + " renamed.";

                    message += " (" + reverseTotalDirectoriesProcessed + "/" + directoriesRemoved.Count + ")";
                }

                if(buttonAbort.Enabled == true)
                {
                    labelStatus.Invoke((MethodInvoker)delegate { labelStatus.Text = message; });
                    labelStatus.Invoke((MethodInvoker)delegate { labelStatus.Update(); });
                }
                else
                {
                    return;
                }
            }
        }

        private void buttonDirectories_Click(object sender, EventArgs e)
        {
            Form form = new FormDirectories();
            form.ShowDialog();

            updateDirectories();
            updateGlobalSettings();

            if (_updateTimerRunning == true)
            {
                _updateTimer.Dispose();
                _updateTimerRunning = false;

                labelNextBackup.Text = "Next backup: Not scheduled";
            }

            if (Properties.Settings.Default.schedule[0] == "enabled" && sources.Count != 0)
            {
                _updateTimer = new System.Threading.Timer(updateTimer, null, 1000, 1000);
                _updateTimerRunning = true;
            }
        }

        private void buttonSchedule_Click(object sender, EventArgs e)
        {
            Form form = new FormSchedule();
            form.ShowDialog();

            if (buttonBackupNow.Enabled == false) return;

            if (_updateTimerRunning == true)
            {
                _updateTimer.Dispose();
                _updateTimerRunning = false;

                labelNextBackup.Text = "Next backup: Not scheduled";
            }

            if (Properties.Settings.Default.schedule[0] == "enabled" && sources.Count != 0)
            {
                _updateTimer = new System.Threading.Timer(updateTimer, null, 1000, 1000);
                _updateTimerRunning = true;
            }
        }

        private void buttonSettings_Click(object sender, EventArgs e)
        {
            Form form = new FormSettings();
            form.ShowDialog();

            updateGlobalSettings();
        }

        private void buttonLog_Click(object sender, EventArgs e)
        {
            if (File.Exists(logFilePath) == true) Process.Start(logFilePath);
            else
            {
                MessageBox.Show("Log file not found");
            }
        }

        private void buttonWarnings_Click(object sender, EventArgs e)
        {
            formWarnings = new FormWarnings();
            formWarnings.ShowDialog();

            formWarnings = null;
        }

        void updateTimer(object state)
        {
            if(nextUpdateDate == new DateTime()) nextUpdateDate = calculateTimer();

            DateTime currentDate = DateTime.Now;
            TimeSpan nextTime = nextUpdateDate - currentDate;

            string nextBackupMessage;

            /*if(nextTime.TotalSeconds < 60) nextBackupMessage = "Next backup: " + nextTime.Seconds + " seconds";
            else if(nextTime.TotalMinutes < 60) nextBackupMessage = "Next backup: " + nextTime.Minutes + " minutes, " + nextTime.Seconds + " seconds";
            else if(nextTime.TotalHours < 24) nextBackupMessage = "Next backup: " + nextTime.Hours + " hours, " + nextTime.Minutes + " minutes, " + nextTime.Seconds + " seconds";
            else nextBackupMessage = "Next backup: " + nextTime.Days + " days, " + nextTime.Hours + " hours, " + nextTime.Minutes + " minutes, " + nextTime.Seconds + " seconds";*/

            nextBackupMessage = nextTime.Seconds == 1 ? nextTime.Seconds + " second" : nextTime.Seconds + " seconds";
            if (nextTime.TotalSeconds > 60) nextBackupMessage = nextTime.Minutes == 1 ? nextTime.Minutes + " minute, " + nextBackupMessage : nextTime.Minutes + " minutes, " + nextBackupMessage;
            if (nextTime.TotalMinutes > 60) nextBackupMessage = nextTime.Hours == 1 ? nextTime.Hours + " hour, " + nextBackupMessage : nextTime.Hours + " hours, " + nextBackupMessage;
            if (nextTime.TotalHours > 24) nextBackupMessage = nextTime.Days == 1 ? nextTime.Days + " day, " + nextBackupMessage : nextTime.Days + " days, " + nextBackupMessage;
            nextBackupMessage = "Next backup: " + nextBackupMessage;

            labelNextBackup.Invoke((MethodInvoker)delegate { labelNextBackup.Text = nextBackupMessage; });

            if (nextTime.TotalSeconds < 0)
            {
                nextUpdateDate = new DateTime();
                backup(true);
            }
        }

        void threeDotTimer(object state)
        {
            if (labelThreeDot.Text == "Status:") labelThreeDot.Invoke((MethodInvoker)delegate { labelThreeDot.Text = "Status."; });
            else if (labelThreeDot.Text == "Status.") labelThreeDot.Invoke((MethodInvoker)delegate { labelThreeDot.Text = "Status.."; });
            else if (labelThreeDot.Text == "Status..") labelThreeDot.Invoke((MethodInvoker)delegate { labelThreeDot.Text = "Status..."; });
            else if (labelThreeDot.Text == "Status...") labelThreeDot.Invoke((MethodInvoker)delegate { labelThreeDot.Text = "Status."; });
            labelThreeDot.Invoke((MethodInvoker)delegate { labelThreeDot.Update(); });
        }

        /*
         * schedule[0] = enabled/disabled
         * schedule[1] = Minutely/Hourly/Daily/Weekly/Monthly
         * schedule[2] = (hour)
         * schedule[3] = (minute)
         * schedule[4] = AM/PM
         * schedule[5] = Sunday/Monday/Tuesday/Wednesday/Thursday/Friday/Saturday
         * schedule[6] = (day of month)
         */

        DateTime calculateTimer()
        {
            DateTime currentDate = DateTime.Now;
            DateTime targetDate = new DateTime();

            int savedMinute = Convert.ToInt32(Properties.Settings.Default.schedule[3]);
            int savedHour = Convert.ToInt32(Properties.Settings.Default.schedule[2]);

            if (Properties.Settings.Default.schedule[4] == "AM" && savedHour == 12) savedHour = 0;
            if (Properties.Settings.Default.schedule[4] == "PM" && savedHour != 12) savedHour += 12;

            if (Properties.Settings.Default.schedule[1] == "Minutely")
            {
                targetDate = new DateTime(currentDate.Year, currentDate.Month, currentDate.Day, currentDate.Hour, currentDate.Minute, 0);
                targetDate = targetDate.AddMinutes(1);

                if (targetDate <= currentDate) targetDate = targetDate.AddHours(1);
            }

            if (Properties.Settings.Default.schedule[1] == "Hourly")
            {
                targetDate = new DateTime(currentDate.Year, currentDate.Month, currentDate.Day, currentDate.Hour, savedMinute, 0);
                if (targetDate <= currentDate) targetDate = targetDate.AddHours(1);
                if (targetDate <= currentDate) targetDate = targetDate.AddDays(1);
            }

            if (Properties.Settings.Default.schedule[1] == "Daily")
            {
                targetDate = new DateTime(currentDate.Year, currentDate.Month, currentDate.Day, savedHour, savedMinute, 0);
                if (targetDate <= currentDate) targetDate = targetDate.AddDays(1);
            }

            if (Properties.Settings.Default.schedule[1] == "Weekly")
            {
                int savedDayOfWeek = Array.IndexOf(daysOfWeek, Properties.Settings.Default.schedule[5]) + 1;
                int currentDayOfWeek = Array.IndexOf(daysOfWeek, DateTime.Today.DayOfWeek.ToString()) + 1;
                int daysIntil;

                if (savedDayOfWeek >= currentDayOfWeek) daysIntil = savedDayOfWeek - currentDayOfWeek;
                else daysIntil = savedDayOfWeek - currentDayOfWeek + 7;

                targetDate = new DateTime(currentDate.Year, currentDate.Month, currentDate.Day, savedHour, savedMinute, 0);
                targetDate = targetDate.AddDays(daysIntil);
                if (targetDate <= currentDate) targetDate = targetDate.AddDays(7);
            }

            if (Properties.Settings.Default.schedule[1] == "Monthly")
            {
                int savedDayOfMonth = Convert.ToInt32(Properties.Settings.Default.schedule[6]);

                if (savedDayOfMonth > DateTime.DaysInMonth(currentDate.Year, currentDate.Month)) savedDayOfMonth = DateTime.DaysInMonth(currentDate.Year, currentDate.Month);

                targetDate = new DateTime(currentDate.Year, currentDate.Month, savedDayOfMonth, savedHour, savedMinute, 0);
                if (targetDate <= currentDate) targetDate = targetDate.AddMonths(1);
                if (targetDate <= currentDate) targetDate = targetDate.AddDays(1);
            }

            return targetDate;
        }

        void updateDirectories()
        {
            listDirectories.Items.Clear();

            for (int i = 0; i < sources.Count(); i++)
            {
                ListViewItem listItem = new ListViewItem((i + 1).ToString());
                listItem.SubItems.Add(sources[i]);
                listItem.SubItems.Add(destinations[i]);
                listDirectories.Items.Add(listItem);
            }
        }

        void updateGlobalSettings()
        {
            if (Properties.Settings.Default.updated == "overwrite") labelUpdated.Text = "Updated files will be overwritten";
            else if (Properties.Settings.Default.updated == "rename") labelUpdated.Text = "Updated files will be renamed";

            if (Properties.Settings.Default.removed == "delete") labelRemoved.Text = "Removed files/folders will be deleted";
            else if (Properties.Settings.Default.removed == "rename") labelRemoved.Text = "Removed files/folders will be renamed";

            bool exceptionsInEffect = false;
            if (exceptions != null) for (int i = 0; i < exceptions.Length; i++) if(exceptions[i][0].ToString() != "(null)") exceptionsInEffect = true;

            if (exceptionsInEffect == true) labelExceptions.Text = "Exceptions in effect";
            else labelExceptions.Text = "Exceptions not in effect";
        }

        void logLog(string log)
        {
            DateTime currentDateTime = DateTime.Now;
            string formattedDateTime = currentDateTime.ToString("dd/MM/yyyy - HH:mm:ss");

            using (StreamWriter outputFile = new StreamWriter(logFilePath, true))
            {
                outputFile.WriteLine("[" + formattedDateTime + "] " + log);
            }
        }

        void logWarning(string warning)
        {
            DateTime currentDateTime = DateTime.Now;
            string formattedDateTime = currentDateTime.ToString("dd/MM/yyyy - HH:mm:ss");

            warnings[0].Insert(0, formattedDateTime);
            warnings[1].Insert(0, warning);
            buttonWarnings.Invoke((MethodInvoker)delegate { buttonWarnings.Text = "Warnings (" + warnings[0].Count() + ")"; });
            buttonWarnings.Invoke((MethodInvoker)delegate { buttonWarnings.Update(); });

            if (formWarnings != null) formWarnings.Invoke((MethodInvoker)delegate { formWarnings.insertNewWarning(); });
        }

        private void clearHistoryToolStripMenuItem_Click(object sender, EventArgs e)
        {
            history[0].Clear();
            history[1].Clear();

            if (Properties.Settings.Default.historyTime != null) Properties.Settings.Default.historyTime.Clear();
            if (Properties.Settings.Default.historyMessage != null) Properties.Settings.Default.historyMessage.Clear();
            Properties.Settings.Default.Save();

            listActions.Items.Clear();
        }

        private void defaultThemeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Properties.Settings.Default.theme = "default";
            Properties.Settings.Default.Save();

            this.BackColor = SystemColors.Control;
            Themes.changeTheme("default", this.Controls);

            colorRed = Color.DarkRed;
            colorGreen = Color.DarkGreen;
            colorBlack = SystemColors.ControlText;
        }

        private void darkModeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Properties.Settings.Default.theme = "dark";
            Properties.Settings.Default.Save();

            this.BackColor = Color.FromArgb(26, 26, 27);
            Themes.changeTheme("dark", this.Controls);

            colorRed = Color.Red;
            colorGreen = Color.LightGreen;
            colorBlack = Color.White;
        }

        private void localBackupUtilityToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Form form = new FormAbout();
            form.ShowDialog();
        }

        private void checkForUpdateToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                HttpWebRequest request = (HttpWebRequest)WebRequest.Create("http://www.lycogala.com/software/LocalBackupUtility.txt");
                request.Timeout = 1000;

                using (HttpWebResponse response = (HttpWebResponse)request.GetResponse())
                using (Stream stream = response.GetResponseStream())
                using (StreamReader reader = new StreamReader(stream))
                {
                    string text = reader.ReadToEnd();
                    string[] lines = text.Split(new[] { '\r', '\n' }, StringSplitOptions.RemoveEmptyEntries);

                    bool newVersion = false;
                    if (versionLocalBackupUtility < float.Parse(lines[0])) newVersion = true;

                    if (newVersion == true)
                    {
                        Form form = new FormUpdate();
                        form.ShowDialog();
                    }
                    else
                    {
                        MessageBox.Show("Local Backup Utility is at the latest version");
                    }
                }
            }
            catch (Exception)
            {
                MessageBox.Show("Unable to check for update at this time" + Environment.NewLine + "Please try again later");
            }
        }

        private void listDirectories_DoubleClick(object sender, EventArgs e)
        {
            buttonDirectories.PerformClick();
        }

        private void listDirectories_MouseClick(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right && buttonBackupNow.Enabled == true)
            {
                var focusedItem = listDirectories.FocusedItem;
                if (focusedItem != null && focusedItem.Bounds.Contains(e.Location))
                {
                    contextMenuStrip1.Show(Cursor.Position);
                }
            }
        }

        private void backupToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Thread thread = new Thread(() => backup(false));
            thread.Start();
        }

        private void directoriesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            buttonDirectories.PerformClick();
        }

        private void scheduleToolStripMenuItem_Click(object sender, EventArgs e)
        {
            buttonSchedule.PerformClick();
        }

        private void globalSettingsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            buttonSettings.PerformClick();
        }

        private void logToolStripMenuItem_Click(object sender, EventArgs e)
        {
            buttonLog.PerformClick();
        }

        private void warningsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            buttonWarnings.PerformClick();
        }

        private void labelUpdatedRemoved_DoubleClick(object sender, EventArgs e)
        {
            buttonSettings.PerformClick();
        }

        private void exceptionsDebugToolStripMenuItem_Click(object sender, EventArgs e)
        {
            bool checkSources = object.ReferenceEquals(sources, null);
            bool checkDestnations = object.ReferenceEquals(destinations, null);
            bool checkExceptions = object.ReferenceEquals(exceptions, null);

            string message;

            if (checkSources == true || checkDestnations == true || checkExceptions == true)
            {
                message = "Sources null: " + checkSources + Environment.NewLine + "Destinations null: " + checkDestnations + Environment.NewLine + "Exceptions null: " + checkExceptions;
            }
            else
            {
                message = "Sources null: " + checkSources + Environment.NewLine + "Destinations null: " + checkDestnations + Environment.NewLine + "Exceptions null: " + checkExceptions +
                     Environment.NewLine + Environment.NewLine +
                     "Number of sources: " + sources.Count + Environment.NewLine + "Number of destinations: " + destinations.Count + Environment.NewLine + "Number of exceptions: " + exceptions.Length;

                if (exceptions.Length > 0)
                {
                    message = message + Environment.NewLine;

                    for (int i = 0; i < exceptions.Length; i++)
                    {
                        message = message + Environment.NewLine + "Exception " + i + ": " + exceptions[i].Count();
                    }
                }
            }

            MessageBox.Show(message);
        }

        private void scheduleDebugToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Form form = new FormCalculateTimerDebug();
            form.ShowDialog();
        }
    }
}
