﻿using System;
using System.Data;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.ComponentModel;
using System.Threading.Tasks;
using System.Collections.Generic;

namespace Local_Backup_Utility
{
    public partial class FormSchedule : Form
    {
        bool ready = false;

        public FormSchedule()
        {
            InitializeComponent();
        }

        /*
         * schedule[0] = enabled/disabled
         * schedule[1] = Minutely/Hourly/Daily/Weekly/Monthly
         * schedule[2] = (hour)
         * schedule[3] = (minute)
         * schedule[4] = AM/PM
         * schedule[5] = Sunday/Monday/Tuesday/Wednesday/Thursday/Friday/Saturday
         * schedule[6] = (day of month)
         */

        private void FormSchedule_Load(object sender, EventArgs e)
        {
            if (Properties.Settings.Default.theme == "dark")
            {
                this.BackColor = Color.FromArgb(26, 26, 27);
                Themes.changeTheme("dark", this.Controls);
            }

            if (Properties.Settings.Default.schedule[0] == "enabled") radioEnabled.Checked = true;
            else if (Properties.Settings.Default.schedule[0] == "disabled") radioDisabled.Checked = true;

            comboFrequency.SelectedItem = Properties.Settings.Default.schedule[1];

            comboHour.SelectedItem = Properties.Settings.Default.schedule[2];
            comboMinute.SelectedItem = Properties.Settings.Default.schedule[3];
            if (Properties.Settings.Default.schedule[4] == "AM") radioAM.Checked = true;
            else if (Properties.Settings.Default.schedule[4] == "PM") radioPM.Checked = true;

            comboDayOfWeek.SelectedItem = Properties.Settings.Default.schedule[5];
            comboDayOfMonth.SelectedItem = Properties.Settings.Default.schedule[6];

            ready = true;
            configureControls();
        }

        private void comboFrequency_SelectedIndexChanged(object sender, EventArgs e)
        {
            configureControls();
        }

        void configureControls()
        {
            if (ready == false) return;

            if(comboFrequency.SelectedItem.ToString() == "Minutely")
            {
                comboHour.Enabled = false;
                comboMinute.Enabled = false;
                radioAM.Enabled = false;
                radioPM.Enabled = false;
                comboDayOfWeek.Enabled = false;
                comboDayOfMonth.Enabled = false;
                labelDayOfMonth.Visible = false;

                groupTime.Visible = false;
                groupDayOfWeek.Visible = false;
                groupDayOfMonth.Visible = false;
            }

            if (comboFrequency.SelectedItem.ToString() == "Hourly")
            {
                comboHour.Enabled = false;
                comboMinute.Enabled = true;
                radioAM.Enabled = false;
                radioPM.Enabled = false;
                comboDayOfWeek.Enabled = false;
                comboDayOfMonth.Enabled = false;
                labelDayOfMonth.Visible = false;

                groupTime.Visible = true;
                groupDayOfWeek.Visible = false;
                groupDayOfMonth.Visible = false;
            }

            if (comboFrequency.SelectedItem.ToString() == "Daily")
            {
                comboHour.Enabled = true;
                comboMinute.Enabled = true;
                radioAM.Enabled = true;
                radioPM.Enabled = true;
                comboDayOfWeek.Enabled = false;
                comboDayOfMonth.Enabled = false;
                labelDayOfMonth.Visible = false;

                groupTime.Visible = true;
                groupDayOfWeek.Visible = false;
                groupDayOfMonth.Visible = false;
            }

            if (comboFrequency.SelectedItem.ToString() == "Weekly")
            {
                comboHour.Enabled = true;
                comboMinute.Enabled = true;
                radioAM.Enabled = true;
                radioPM.Enabled = true;
                comboDayOfWeek.Enabled = true;
                comboDayOfMonth.Enabled = false;
                labelDayOfMonth.Visible = false;

                groupTime.Visible = true;
                groupDayOfWeek.Visible = true;
                groupDayOfMonth.Visible = false;
            }

            if (comboFrequency.SelectedItem.ToString() == "Monthly")
            {
                comboHour.Enabled = true;
                comboMinute.Enabled = true;
                radioAM.Enabled = true;
                radioPM.Enabled = true;
                comboDayOfWeek.Enabled = false;
                comboDayOfMonth.Enabled = true;
                labelDayOfMonth.Visible = true;

                groupTime.Visible = true;
                groupDayOfWeek.Visible = false;
                groupDayOfMonth.Visible = true;
            }

            if (Properties.Settings.Default.theme == "dark")
            {
                this.BackColor = Color.FromArgb(26, 26, 27);
                Themes.changeTheme("dark", this.Controls);
            }
        }

        private void buttonSaveClose_Click(object sender, EventArgs e)
        {
            if (radioEnabled.Checked == true) Properties.Settings.Default.schedule[0] = "enabled";
            else if (radioDisabled.Checked == true ) Properties.Settings.Default.schedule[0] = "disabled";

            Properties.Settings.Default.schedule[1] = comboFrequency.SelectedItem.ToString();

            Properties.Settings.Default.schedule[2] = comboHour.SelectedItem.ToString();
            Properties.Settings.Default.schedule[3] = comboMinute.SelectedItem.ToString();
            if (radioAM.Checked == true) Properties.Settings.Default.schedule[4] = "AM";
            else if (radioPM.Checked == true) Properties.Settings.Default.schedule[4] = "PM";

            Properties.Settings.Default.schedule[5] = comboDayOfWeek.SelectedItem.ToString();
            Properties.Settings.Default.schedule[6] = comboDayOfMonth.SelectedItem.ToString();

            Properties.Settings.Default.Save();
            this.Close();
        }

        private void FormSchedule_FormClosing(object sender, FormClosingEventArgs e)
        {
            bool unsavedChanges = false;

            if (radioEnabled.Checked == true && Properties.Settings.Default.schedule[0] != "enabled") unsavedChanges = true;
            if (radioDisabled.Checked == true && Properties.Settings.Default.schedule[0] != "disabled") unsavedChanges = true;

            if (Properties.Settings.Default.schedule[1] != comboFrequency.SelectedItem.ToString()) unsavedChanges = true;

            if (Properties.Settings.Default.schedule[2] != comboHour.SelectedItem.ToString()) unsavedChanges = true;
            if (Properties.Settings.Default.schedule[3] != comboMinute.SelectedItem.ToString()) unsavedChanges = true;
            if (radioAM.Checked == true && Properties.Settings.Default.schedule[4] != "AM") unsavedChanges = true;
            if (radioPM.Checked == true && Properties.Settings.Default.schedule[4] != "PM") unsavedChanges = true;

            if(Properties.Settings.Default.schedule[5] != comboDayOfWeek.SelectedItem.ToString()) unsavedChanges = true;
            if(Properties.Settings.Default.schedule[6] != comboDayOfMonth.SelectedItem.ToString()) unsavedChanges = true;

            if (unsavedChanges == true)
            {
                DialogResult checkExit = MessageBox.Show("You have unsaved changes" + Environment.NewLine + "Are you sure you want to close this form?", "Unsaved changes", MessageBoxButtons.YesNo);

                if (checkExit == DialogResult.No) e.Cancel = true;
            }
        }
    }
}
