﻿using System;
using System.Data;
using System.Linq;
using System.Text;
using System.Drawing;
using Microsoft.Win32;
using System.Windows.Forms;
using System.ComponentModel;
using System.Threading.Tasks;
using System.Collections.Generic;
using System.IO;

namespace Local_Backup_Utility
{
    public partial class FormSettings : Form
    {
        public FormSettings()
        {
            InitializeComponent();
        }

        private void FormGlobalSettings_Load(object sender, EventArgs e)
        {
            if (Properties.Settings.Default.theme == "dark")
            {
                this.BackColor = Color.FromArgb(26, 26, 27);
                Themes.changeTheme("dark", this.Controls);

                labelWarningDeleted.ForeColor = Color.Red;
                labelWarningOverwritten.ForeColor = Color.Red;
            }

            if (Properties.Settings.Default.updated == "overwrite") radioUpdatedOverwrite.Checked = true;
            else if (Properties.Settings.Default.updated == "rename") radioUpdatedRename.Checked = true;

            if (Properties.Settings.Default.removed == "delete") radioRemovedDelete.Checked = true;
            else if (Properties.Settings.Default.removed == "rename") radioRemovedRename.Checked = true;

            if (Properties.Settings.Default.startStartup == true) checkStartStartup.Checked = true;
            else if (Properties.Settings.Default.startStartup == false) checkStartStartup.Checked = false;

            if (Properties.Settings.Default.startMinimised == true) checkStartMinimised.Checked = true;
            else if (Properties.Settings.Default.startMinimised == false) checkStartMinimised.Checked = false;
        }

        private void buttonSaveClose_Click(object sender, EventArgs e)
        {
            if (radioUpdatedOverwrite.Checked == true) Properties.Settings.Default.updated = "overwrite";
            else if (radioUpdatedRename.Checked == true) Properties.Settings.Default.updated = "rename"; 

            if (radioRemovedDelete.Checked == true) Properties.Settings.Default.removed = "delete";
            else if (radioRemovedRename.Checked == true) Properties.Settings.Default.removed = "rename";

            if (checkStartStartup.Checked == true) Properties.Settings.Default.startStartup = true;
            else if (checkStartStartup.Checked == false) Properties.Settings.Default.startStartup = false;

            if (checkStartMinimised.Checked == true) Properties.Settings.Default.startMinimised = true;
            else if (checkStartMinimised.Checked == false) Properties.Settings.Default.startMinimised = false;

            string key = "LocalBackupUtility";
            string executablePath = Application.ExecutablePath;
            RegistryKey keyPath = Registry.CurrentUser.OpenSubKey("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run", true);

            if (checkStartStartup.Checked == true) keyPath.SetValue(key, executablePath);
            else keyPath.DeleteValue(key, false);

            /*string shortcutPath = Environment.GetFolderPath(Environment.SpecialFolder.Startup) + "\\" + "Local Backup Utility.url";
            string localBackupUtilityPath = System.Reflection.Assembly.GetExecutingAssembly().Location;

            if (checkStartStartup.Checked == true)
            {
                if (File.Exists(shortcutPath) == false)
                {
                    using (StreamWriter writer = new StreamWriter(shortcutPath))
                    {
                        writer.WriteLine("[InternetShortcut]");
                        writer.WriteLine("URL=file:///" + localBackupUtilityPath);
                    }
                }
            }
            else
            {
                if(File.Exists(shortcutPath) == true)
                {
                    File.Delete(shortcutPath);
                }
            }*/

            Properties.Settings.Default.Save();
            this.Close();
        }

        private void FormGlobalSettings_FormClosing(object sender, FormClosingEventArgs e)
        {
            bool unsavedChanges = false;

            if (Properties.Settings.Default.updated == "overwrite" && radioUpdatedOverwrite.Checked != true) unsavedChanges = true;
            if (Properties.Settings.Default.updated == "rename" && radioUpdatedRename.Checked != true) unsavedChanges = true;

            if (Properties.Settings.Default.removed == "delete" && radioRemovedDelete.Checked != true) unsavedChanges = true;
            if (Properties.Settings.Default.removed == "rename" && radioRemovedRename.Checked != true) unsavedChanges = true;

            if (Properties.Settings.Default.startStartup == true && checkStartStartup.Checked != true) unsavedChanges = true;
            if (Properties.Settings.Default.startStartup == false && checkStartStartup.Checked != false) unsavedChanges = true;

            if (Properties.Settings.Default.startMinimised == true && checkStartMinimised.Checked != true) unsavedChanges = true;
            if (Properties.Settings.Default.startMinimised == false && checkStartMinimised.Checked != false) unsavedChanges = true;

            if (unsavedChanges == true)
            {
                DialogResult checkExit = MessageBox.Show("You have unsaved changes" + Environment.NewLine + "Are you sure you want to close this form?", "Unsaved changes", MessageBoxButtons.YesNo);

                if (checkExit == DialogResult.No) e.Cancel = true;
            }
        }
    }
}
