﻿using System;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.Threading.Tasks;
using System.Collections.Generic;

namespace Local_Backup_Utility
{
    class Themes
    {
        public static void changeTheme(string theme, Control.ControlCollection container)
        {
            if (theme == "default")
            {
                foreach (Control component in container)
                {
                    component.ForeColor = default(Color);
                    component.BackColor = default(Color);

                    if (component is Button)
                    {
                        ((Button)component).FlatStyle = FlatStyle.Standard;
                        ((Button)component).UseVisualStyleBackColor = true;
                    }

                    if (component is MenuStrip menuStrip) menuStrip.Renderer = null;

                    if (component is ListView listView)
                    {
                        listView.OwnerDraw = false;
                        listView.BorderStyle = BorderStyle.Fixed3D;
                    }

                    if (component is GroupBox) changeTheme("default", component.Controls);
                }
            }

            if (theme == "dark")
            {
                foreach (Control component in container)
                {
                    if (component is Label)
                    {
                        component.ForeColor = Color.White;
                        component.BackColor = default(Color);
                    }

                    if (component is LinkLabel linkLabel)
                    {
                        linkLabel.LinkColor = Color.White;
                    }

                    if (component is RadioButton)
                    {
                        component.ForeColor = Color.White;
                        component.BackColor = default(Color);
                    }

                    if (component is CheckBox)
                    {
                        component.ForeColor = Color.White;
                        component.BackColor = default(Color);
                    }

                    if (component is TextBox)
                    {
                        component.ForeColor = Color.White;
                        component.BackColor = Color.FromArgb(52, 52, 54);
                    }

                    if (component is Button button)
                    {
                        button.ForeColor = Color.White;
                        button.BackColor = Color.FromArgb(26, 26, 27);

                        button.FlatStyle = FlatStyle.Flat;
                    }

                    if (component is ComboBox comboBox)
                    {
                        comboBox.ForeColor = Color.White;
                        comboBox.BackColor = Color.FromArgb(52, 52, 54);

                        comboBox.DrawMode = DrawMode.OwnerDrawFixed;
                        comboBox.DrawItem += (sender, e) => ComboBox_DrawItem(sender, e, Color.White, Color.FromArgb(52, 52, 54));
                    }

                    if (component is MenuStrip menuStrip)
                    {
                        ColorizeMenuStrip(ref menuStrip, Color.FromArgb(26, 26, 27), Color.White, Color.FromArgb(52, 52, 54));
                    }

                    if (component is GroupBox)
                    {
                        component.ForeColor = Color.White;
                        component.BackColor = Color.FromArgb(26, 26, 27);

                        changeTheme("dark", component.Controls);
                    }

                    if (component is ListView listView)
                    {
                        listView.BackColor = Color.FromArgb(52, 52, 54);
                        listView.ForeColor = Color.White;

                        listView.OwnerDraw = true;
                        listView.DrawColumnHeader += (sender, e) => ListView_DrawColumnHeader(sender, e, Color.White, Color.FromArgb(52, 52, 54));
                        listView.DrawItem += ListView_DrawItem;

                        listView.BorderStyle = BorderStyle.FixedSingle;
                    }

                    if (component is ListBox)
                    {
                        component.ForeColor = Color.White;
                        component.BackColor = Color.FromArgb(52, 52, 54);
                    }

                    if (component.Enabled == false) component.BackColor = Color.Gray;
                }
            }
        }

        /* 
         * THE FOLLOWING WAS CODED BY CHATGPT
         */

        //--- Combo Box ---//
        //--- Working ---//

        private static void ComboBox_DrawItem(object sender, DrawItemEventArgs e, Color foreColor, Color backColor)
        {
            if (e.Index < 0)
                return;

            e.Graphics.FillRectangle(new SolidBrush(backColor), e.Bounds);

            e.Graphics.DrawString(((ComboBox)sender).Items[e.Index].ToString(), e.Font, new SolidBrush(foreColor), e.Bounds, StringFormat.GenericDefault);

            if ((e.State & DrawItemState.Selected) == DrawItemState.Selected)
            {
                using (Pen focusPen = new Pen(Color.White))
                {
                    e.Graphics.DrawRectangle(focusPen, e.Bounds.X, e.Bounds.Y, e.Bounds.Width - 1, e.Bounds.Height - 1);
                }
            }
        }

        //--- List View ---//
        //--- Almost Working ---//

        private static void ListView_DrawColumnHeader(object sender, DrawListViewColumnHeaderEventArgs e, Color foreColor, Color backColor)
        {
            e.Graphics.FillRectangle(new SolidBrush(backColor), e.Bounds);

            Rectangle textBounds = e.Bounds;
            textBounds.X += 3;
            textBounds.Y += 5;

            e.Graphics.DrawString(e.Header.Text, e.Font, new SolidBrush(foreColor), textBounds, StringFormat.GenericDefault);
        }

        private static void ListView_DrawItem(object sender, DrawListViewItemEventArgs e)
        {
            e.DrawDefault = true;
        }

        //--- Menu Strip ---//
        //--- Working ---//

        public static void ColorizeMenuStrip(ref MenuStrip menuStrip, Color backColor, Color foreColor, Color selectionColor)
        {
            menuStrip.RenderMode = ToolStripRenderMode.Professional;
            menuStrip.Renderer = new CustomMenuStripRenderer(backColor, foreColor, selectionColor);
        }

        private class CustomMenuStripRenderer : ToolStripProfessionalRenderer
        {
            private readonly Color _backColor;
            private readonly Color _foreColor;
            private readonly Color _selectionColor;

            public CustomMenuStripRenderer(Color backColor, Color foreColor, Color selectionColor)
            {
                _backColor = backColor;
                _foreColor = foreColor;
                _selectionColor = selectionColor;
            }

            protected override void OnRenderItemText(ToolStripItemTextRenderEventArgs e)
            {
                e.TextColor = _foreColor;
                base.OnRenderItemText(e);
            }

            protected override void OnRenderToolStripBackground(ToolStripRenderEventArgs e)
            {
                using (SolidBrush backBrush = new SolidBrush(_backColor))
                {
                    e.Graphics.FillRectangle(backBrush, e.AffectedBounds);
                }
            }

            protected override void OnRenderMenuItemBackground(ToolStripItemRenderEventArgs e)
            {
                Rectangle bounds = new Rectangle(Point.Empty, e.Item.Size);
                Color fillColor = e.Item.Selected ? _selectionColor : e.Item.Bounds.Contains(e.ToolStrip.PointToClient(Control.MousePosition)) ? _selectionColor : _backColor;

                using (SolidBrush fillBrush = new SolidBrush(fillColor))
                {
                    e.Graphics.FillRectangle(fillBrush, bounds);
                }
            }

            protected override void OnRenderImageMargin(ToolStripRenderEventArgs e)
            {
                // Override to avoid rendering the image margin
            }

            protected override void OnRenderToolStripBorder(ToolStripRenderEventArgs e)
            {
                // Override to avoid rendering the border
            }
        }
    }
}
